"""Layers of a convolutional encoder-decoder network."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-

# externals
import torch.nn as nn
import torch.nn.functional as F

# locals
from pysegcnn.core.layers import Conv2dSame, EncoderBlock, DecoderBlock


def conv_bn_relu(in_channels, out_channels, **kwargs):
    """Block of convolution, batch normalization and rectified linear unit.

    Parameters
    ----------
    in_channels : `int`
        Number of input channels.
    out_channels : `int`
        Number of output channels.
    **kwargs: `dict` [`str`]
        Additional arguments passed to `pysegcnn.core.layers.Conv2dSame`.

    Returns
    -------
    block : `torch.nn.Sequential` [`torch.nn.Module`]
        An instance of `torch.nn.Sequential` containing a sequence of
        convolution, batch normalization and rectified linear unit layers.

    """
    return nn.Sequential(
            Conv2dSame(in_channels, out_channels, **kwargs),
            nn.BatchNorm2d(out_channels),
            nn.ReLU())


class ConvBnReluConvS2(EncoderBlock):
    """Block of convolution, batchnorm, relu and convolution with stride=2."""

    def __init__(self, in_channels, out_channels, **kwargs):
        super().__init__(in_channels, out_channels, **kwargs)

        # downsampling layer: convolution with stride=2
        self.pool = conv_bn_relu(out_channels, out_channels, kernel_size=3,
                                 stride=2)

    def layers(self):
        """Sequence of convolution, batchnorm and relu layers.

        Returns
        -------
        layers : :py:class:`torch.nn.Sequential` [:py:class:`torch.nn.Module`]
            An instance of :py:class:`torch.nn.Sequential` containing the
            sequence of convolution, batchnorm and relu layer
            (:py:class:`torch.nn.Module`) instances.

        """
        return conv_bn_relu(self.in_channels, self.out_channels, **self.kwargs)

    def downsample(self, x):
        """Convolutional layer with stride=2.

        Parameters
        ----------
        x : :py:class:`torch.Tensor`, shape=(b, c, h, w)
            Input tensor.

        Returns
        -------
        x : :py:class:`torch.Tensor`, shape=(b, c, h // 2, w // 2)
            The 2x2 strided tensor.
        indices : :py:class:`torch.Tensor` or `None`
            The indices of the maxima. Useful for upsampling with
            :py:func:`torch.nn.functional.max_unpool2d`.

        """
        return self.pool(x), None


class ConvBnReluUpsample(DecoderBlock):
    """Block of convolution, batchnorm, relu and nearest neighbor upsample."""

    def __init__(self, in_channels, out_channels, **kwargs):
        super().__init__(in_channels, out_channels, **kwargs)

    def layers(self):
        """Sequence of convolution, batchnorm and relu layers.

        Returns
        -------
        layers : :py:class:`torch.nn.Sequential` [:py:class:`torch.nn.Module`]
            An instance of :py:class:`torch.nn.Sequential` containing the
            sequence of convolution, batchnorm and relu layer
            (:py:class:`torch.nn.Module`) instances.

        """
        return conv_bn_relu(self.in_channels, self.out_channels, **self.kwargs)

    def upsample(self, x, feature, indices=None):
        """Nearest neighbor upsampling.

        Parameters
        ----------
        x : :py:class:`torch.Tensor`, shape=(b, c, h, w)
            Input tensor.
        feature : :py:class:`torch.Tensor`, shape=(b, c, h, w)
            Intermediate output of a layer in the encoder. Used to determine
            the output shape of the upsampling operation.
        indices : `None`, optional
            The indices of the maxima of the max pooling operation
            (as returned by :py:func:`torch.nn.functional.max_pool2d`).
            Not required by this upsampling method.

        Returns
        -------
        x : :py:class:`torch.Tensor`, shape=(b, c, h, w)
            The 2x2 upsampled tensor.

        """
        return F.interpolate(x, size=feature.shape[2:], mode='nearest')

    def extra_repr(self):
        """Define optional extra information about this module.

        Returns
        -------
        `str`
            Extra representation string.
        """
        return '(upsample): Upsample(mode="nearest")'
