"""Dynamical climate downscaling using deep convolutional neural networks."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-

# builtins
import sys
import time
import logging
from datetime import timedelta
from logging.config import dictConfig

# externals
import xarray as xr

# locals
from pysegcnn.core.trainer import LogConfig
from pysegcnn.core.models import Network
from pysegcnn.core.logging import log_conf
from climax.core.dataset import ERA5Dataset
from climax.core.constants import ERA5_VARIABLES
from climax.core.predict import predict_ERA5
from climax.core.utils import split_date_range
from climax.main.config import (ERA5_PLEVELS, PREDICTAND, NET, VALID_PERIOD,
                                BATCH_SIZE, NORM, DOY, NYEARS)
from climax.main.io import ERA5_PATH, MODEL_PATH, TARGET_PATH

# module level logger
LOGGER = logging.getLogger(__name__)


if __name__ == '__main__':

    # initialize timing
    start_time = time.monotonic()

    # initialize logging
    log_file = MODEL_PATH.joinpath('{}_log.txt'.format(NET.__name__))
    dictConfig(log_conf(log_file))

    # predict reference period
    LogConfig.init_log('Predicting reference period: {}'.format(
        ' - '.join([str(VALID_PERIOD[0]), str(VALID_PERIOD[-1])])))

    # initialize ERA5 predictor dataset
    LogConfig.init_log('Initializing ERA5 predictors.')
    Era5 = ERA5Dataset(ERA5_PATH.joinpath('ERA5'), ERA5_VARIABLES,
                       plevels=ERA5_PLEVELS)
    Era5_ds = Era5.merge()

    # load pretrained model
    state_file = MODEL_PATH.joinpath('_'.join([NET.__name__, PREDICTAND]) +
                                     '.pt')
    if state_file.exists():
        # load pretrained network
        net, _ = Network.load_pretrained_model(state_file, NET)
    else:
        # initialize OBS predictand dataset
        LOGGER.info('{} does not exist.'.format(state_file))
        sys.exit()

    # subset to reference period and predict in NYEAR intervals
    trg_ds = []
    for dates in split_date_range(VALID_PERIOD[0], VALID_PERIOD[-1],
                                  years=NYEARS):
        LogConfig.init_log('Predicting period: {}'.format(
            ' - '.join([str(dates[0]), str(dates[-1])])))
        ref_ds = Era5_ds.sel(time=dates)
        trg_ds.append(predict_ERA5(net, ref_ds, PREDICTAND, normalize=NORM,
                                   batch_size=BATCH_SIZE, doy=DOY))

    # merge predictions for entire validation period
    LOGGER.info('Merging refernce periods ...')
    trg_ds = xr.concat(trg_ds, dim='time')

    # save model predictions as NetCDF file
    target = TARGET_PATH.joinpath(PREDICTAND, net.state_file.name.replace(
        net.state_file.suffix, '.nc'))
    if not target.parent.exists():
        target.parent.mkdir(parents=True, exist_ok=True)
    LOGGER.info('Saving network predictions: {}.'.format(target))
    trg_ds.to_netcdf(target, engine='h5netcdf')

    # log execution time of script
    LogConfig.init_log('Execution time of script {}: {}'
                       .format(__file__, timedelta(seconds=time.monotonic() -
                                                   start_time)))
