"""Dynamical climate downscaling using deep convolutional neural networks."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-

# builtins
import time
import logging
from datetime import timedelta
from logging.config import dictConfig

# externals
import torch
import xarray as xr
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader

# locals
from pysegcnn.core.utils import search_files
from pysegcnn.core.trainer import NetworkTrainer, LogConfig
from pysegcnn.core.models import Network
from pysegcnn.core.logging import log_conf
from climax.core.dataset import ERA5Dataset, NetCDFDataset
from climax.core.constants import ERA5_VARIABLES
from climax.main.config import (ERA5_PLEVELS, PREDICTAND, CALIB_PERIOD, DOY,
                                SHUFFLE, BATCH_SIZE, LR, LAMBDA, NORM,
                                TRAIN_CONFIG, NET, LOSS, FILTERS, OVERWRITE)
from climax.main.io import ERA5_PATH, OBS_PATH, MODEL_PATH

# module level logger
LOGGER = logging.getLogger(__name__)


if __name__ == '__main__':

    # initialize timing
    start_time = time.monotonic()

    # initialize logging
    log_file = MODEL_PATH.joinpath('{}_log.txt'.format(NET.__name__))
    if log_file.exists():
        log_file.unlink()
    dictConfig(log_conf(log_file))

    # initialize downscaling
    LogConfig.init_log('Initializing downscaling for period: {}'.format(
        ' - '.join([str(CALIB_PERIOD[0]), str(CALIB_PERIOD[-1])])))

    # initialize ERA5 predictor dataset
    LogConfig.init_log('Initializing ERA5 predictors.')
    Era5 = ERA5Dataset(ERA5_PATH.joinpath('ERA5'), ERA5_VARIABLES,
                       plevels=ERA5_PLEVELS)
    Era5_ds = Era5.merge()

    # initialize network filename
    state_file = MODEL_PATH.joinpath('_'.join([NET.__name__, PREDICTAND]) +
                                     '.pt')
    if state_file.exists() and not OVERWRITE:
        # load pretrained network
        net, _ = Network.load_pretrained_model(state_file, NET)
    else:
        # initialize OBS predictand dataset
        LogConfig.init_log('Initializing observations for predictand: {}'
                           .format(PREDICTAND))

        # check whether to joinlty train tasmin and tasmax: diurnal variation
        if PREDICTAND == 'tas':
            # calculate diurnal variation
            tasmax = search_files(OBS_PATH.joinpath('tasmax'), '.nc$').pop()
            tasmin = search_files(OBS_PATH.joinpath('tasmin'), '.nc$').pop()
            Obs_ds = (xr.open_dataset(tasmax) - xr.open_dataset(tasmin))
        else:
            # read in-situ gridded observations
            Obs_ds = search_files(OBS_PATH.joinpath(PREDICTAND), '.nc$').pop()
            Obs_ds = xr.open_dataset(Obs_ds)

        # initialize training data
        LogConfig.init_log('Initializing training data.')

        # split calibration period into training and validation period
        train, valid = train_test_split(CALIB_PERIOD, shuffle=False)

        # training and validation dataset
        Era5_train, Obs_train = Era5_ds.sel(time=train), Obs_ds.sel(time=train)
        Era5_valid, Obs_valid = Era5_ds.sel(time=valid), Obs_ds.sel(time=valid)

        # create PyTorch compliant dataset and dataloader instances for model
        # training
        train_ds = NetCDFDataset(Era5_train, Obs_train, normalize=NORM,
                                 doy=DOY)
        valid_ds = NetCDFDataset(Era5_valid, Obs_valid, normalize=NORM,
                                 doy=DOY)
        train_dl = DataLoader(train_ds, batch_size=BATCH_SIZE, shuffle=SHUFFLE,
                              drop_last=False)
        valid_dl = DataLoader(valid_ds, batch_size=BATCH_SIZE, shuffle=SHUFFLE,
                              drop_last=False)

        # instanciate network from scratch
        in_channels = len(Era5_ds.data_vars)
        net = NET(state_file, in_channels + 2 if DOY else in_channels, 1,
                  filters=FILTERS)

        	# initialize optimizer
        optimizer = torch.optim.Adam(net.parameters(), lr=LR,
                                     weight_decay=LAMBDA)

        # initialize network trainer
        trainer = NetworkTrainer(net, optimizer, net.state_file, train_dl,
                                 valid_dl, loss_function=LOSS, **TRAIN_CONFIG)

        # train model
        state = trainer.train()


    # log execution time of script
    LogConfig.init_log('Execution time of script {}: {}'
                       .format(__file__, timedelta(seconds=time.monotonic() -
                                                   start_time)))
